#include "minimal/stdlib.h"
#include "minimal/string.h"
#include "minimal/hexdump.h"

#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/mman.h>

#include <objc/objc-runtime.h>
#include <objc/objc-class.h>

#include <dlfcn.h>
#include <ffi.h>

#include "joc/Cast.h"
#include "joc/Java.h"
#include "joc/Memory.h"
#include "joc/NativeException.h"
#include "joc/Pointer.h"
#include "joc/RuntimeInfo.h"
#include "joc/Selector.h"
#include "joc/Strap.h"
#include "joc/String.h"

#include "sig/ffi_type.h"
#include "sig/parse.h"

@protocol JOCStrap
+ (id) allocWithObject:(jobject)object;
@end

struct meta_strap {
    JNIEnv *joc_env;
    jclass joc_class;
    jmethodID joc_method;
};

struct instance_strap {
    jobject joc_object;
};

struct objc_ivar_list_ {
    int ivar_count;
    struct objc_ivar ivar_list[];
};

static struct objc_ivar_list_ meta_ivars_ = {3, {
    {"joc_env", "^v", sizeof(JNIEnv *)},
    {"joc_class", "^v", sizeof(jclass)},
    {"joc_method", "^v", sizeof(jclass)}
}};

static struct objc_ivar_list_ instance_ivars_ = {1, {
    {"joc_object", "^v", sizeof(jobject)}
}};

static struct objc_ivar_list_ null_ivars_ = {0
};

/* TAG: gc */
id joc_Strap_alloc(id self, SEL sel) {
    JNIEnv *joc_env;
    object_getInstanceVariable(self, "joc_env", (void **) (void *) &joc_env);
    _assert(joc_env != NULL);

    jclass joc_class;
    object_getInstanceVariable(self, "joc_class", (void **) &joc_class);
    _assert(joc_class != NULL);

    jmethodID joc_method;
    object_getInstanceVariable(self, "joc_method", (void **) (void *) &joc_method);
    _assert(joc_method != NULL);

    jobject object = (*(*joc_env)->NewObject)(joc_env, joc_class, joc_method);
    /* XXX: exception might get thrown */
    _assert(object != NULL);

    id cobject = [joc_FromObject(joc_env, object) retain];
    (*(*joc_env)->DeleteLocalRef)(joc_env, object);
    return cobject;
}

/* TAG: gc */
id joc_Strap_allocWithObject(id self, SEL sel, jobject object) {
    JNIEnv *joc_env;
    object_getInstanceVariable(self, "joc_env", (void **) (void *) &joc_env);
    _assert(joc_env != NULL);

    struct objc_super super = {self, self->isa->super_class};
    while ((super.class->info & CLS_JAVA_CLASS) != 0)
        super.class = super.class->super_class;

    id cobject = objc_msgSendSuper(&super, @selector(alloc));
    _assert(cobject != NULL);
    _assert(cobject->isa == self);

    object = (*(*joc_env)->NewWeakGlobalRef)(joc_env, object);
    object_setInstanceVariable(cobject, "joc_object", (void *) object);

    printf("allocWithObject(%p:%s)=%u\n", cobject, cobject->isa->name, [cobject retainCount]);
    return cobject;
}

struct joc_userdata {
    struct sig_signature *signature;
    JNIEnv *env;
    jmethodID method;
    ffi_cif *cif;
    void (*fn)();
};

/* TAG: gc */
void joc_closure(ffi_cif *cif, void *result, void **arguments, void *arg) {
    struct joc_userdata *userdata = (struct joc_userdata *) arg;

    //printf("joc_closure(%p)\n", userdata);

    JNIEnv *env = userdata->env;

    id self = * (id *) arguments[0];

    jobject _this;
    object_getInstanceVariable(self, "joc_object", (void **) &_this);
    _assert(_this != NULL);

    _this = (*(*env)->NewLocalRef)(env, _this);
    _assert(_this != NULL);

    /*SEL sel = * (SEL *) arguments[1];
    printf("self:%p old_this:%p @selector(%s)\n", self, _this, sel_getName(sel));*/

    void *values[userdata->cif->nargs];
    values[0] = &env;
    values[1] = &_this;
    values[2] = &userdata->method;

    apr_pool_t *pool;
    apr_pool_create(&pool, apr_pool_);

    size_t size = 0;

    for (size_t index = 3; index != userdata->cif->nargs; ++index)
        size += userdata->cif->arg_types[index]->size;

    /* XXX: alignment */
    uint8_t data[size];

    size = 0;
    for (size_t index = 3; index != userdata->cif->nargs; ++index) {
        ffi_type *ffi_type = userdata->cif->arg_types[index];
        values[index] = data + size;
        java_copy_to(pool, userdata->signature->elements[index].type, ffi_type, arguments[index - 1], env, values[index]);
        size += ffi_type->size;
    }

    /* XXX: alignment */
    uint8_t value[userdata->cif->rtype->size];
    ffi_call(userdata->cif, userdata->fn, value, values);

    jthrowable error = ((*env)->ExceptionOccurred)(env);
    if (error == NULL)
        java_copy_from(pool, userdata->signature->elements[0].type, userdata->cif->rtype, result, env, value);

    (*(*env)->DeleteLocalRef)(env, _this);
    apr_pool_destroy(pool);

    if (error != NULL)
        @throw joc_FromObject(env, error);
}

struct objc_method_list_ {
    struct objc_method_list *obsolete;
    int method_count;
    struct objc_method method_list[];
};

id joc_Strap_methodSignatureForSelector$(id self, SEL sel, SEL target) {
    printf("methodSignatureForSelector$(%p, %s, %s)\n", self, sel_getName(sel), sel_getName(target));
    struct objc_super super = {self, self->isa->super_class};
    return objc_msgSendSuper(&super, sel, target);
}

jobject joc_static(id self, SEL sel, JNIEnv *env) {
    jclass _this;
    object_getInstanceVariable(self, "joc_class", (void **) &_this);
    return (*(*env)->NewLocalRef)(env, _this);
}

jobject joc_instance(id self, SEL sel, JNIEnv *env) {
    jobject _this;
    object_getInstanceVariable(self, "joc_object", (void **) &_this);
    return (*(*env)->NewLocalRef)(env, _this);
}

/* TAG: gc */
void joc_Strap_release(id self, SEL sel) {
    unsigned count = [self retainCount];
    //printf("release(%p:%s)=%u>%u\n", self, self->isa->name, count, count - 1);

    if (count == 2) {
        JNIEnv *env;
        object_getInstanceVariable(self->isa, "joc_env", (void **) (void *) &env);
        _assert(env != NULL);

        jobject _this;
        object_getInstanceVariable(self, "joc_object", (void **) &_this);
        _assert(_this != NULL);

        object_setInstanceVariable(self, "joc_object", (void *) (*(*env)->NewWeakGlobalRef)(env, _this));
        (*(*env)->DeleteGlobalRef)(env, _this);
    }

    struct objc_super super = {self, self->isa->super_class};
    while ((super.class->info & CLS_JAVA_CLASS) != 0)
        super.class = super.class->super_class;
    objc_msgSendSuper(&super, sel);
}

/* TAG: gc */
id joc_Strap_retain(id self, SEL sel) {
    unsigned count = [self retainCount];
    //printf("retain(%p:%s)=%u>%u\n", self, self->isa->name, count, count + 1);

    if (count == 1) {
        JNIEnv *env;
        object_getInstanceVariable(self->isa, "joc_env", (void **) (void *) &env);
        _assert(env != NULL);

        jweak _this;
        object_getInstanceVariable(self, "joc_object", (void **) &_this);
        _assert(_this != NULL);

        object_setInstanceVariable(self, "joc_object", (void *) (*(*env)->NewGlobalRef)(env, _this));
        (*(*env)->DeleteWeakGlobalRef)(env, _this);
    }

    struct objc_super super = {self, self->isa->super_class};
    while ((super.class->info & CLS_JAVA_CLASS) != 0)
        super.class = super.class->super_class;
    return objc_msgSendSuper(&super, sel);
}

struct objc_method_list_ meta_method = {NULL, 3, {
    //{(SEL) "methodSignatureForSelector:", "@12@0:4:8", (IMP) &joc_Strap_methodSignatureForSelector$},
    {(SEL) "alloc", "@8@0:4", (IMP) &joc_Strap_alloc},
    {(SEL) "allocWithObject:", "@12@0:4^v8", (IMP) &joc_Strap_allocWithObject},
    {(SEL) "_jocStrap:", "@12@0:4^v8", (IMP) &joc_static}
}};

/* it would be more efficient to make _jocStrap only be implemented once */
struct objc_method_list_ instance_method = {NULL, 3, {
    //{(SEL) "methodSignatureForSelector:", "@12@0:4:8", (IMP) &joc_Strap_methodSignatureForSelector$},
    {(SEL) "_jocStrap:", "@12@0:4^v8", (IMP) &joc_instance},
    {(SEL) "release", "v8@0:4", (IMP) &joc_Strap_release},
    {(SEL) "retain", "@8@0:4", (IMP) &joc_Strap_retain}
}};

struct objc_method_list *meta_methods[] = { (struct objc_method_list *) (void *) &meta_method, NULL };
struct objc_method_list *instance_methods[] = { (struct objc_method_list *) (void *) &instance_method, NULL };

void jocSelect_(struct objc_method_list *method_list) {
    for (int index = 0; index != method_list->method_count; ++index) {
        struct objc_method *method = &method_list->method_list[index];
        method->method_name = sel_registerName((char *) method->method_name);
    }
}

void copy_ivars(struct objc_class *new_class, struct objc_ivar_list_ *old_ivars) {
    size_t offset = new_class->super_class->instance_size;
    struct objc_ivar_list *new_ivars = apr_palloc(apr_pool_, sizeof(struct objc_ivar_list_) + old_ivars->ivar_count * sizeof(struct objc_ivar));

    int count = new_ivars->ivar_count = old_ivars->ivar_count;
    for (int ivar = 0; ivar != count; ++ivar) {
        struct objc_ivar *old_ivar = &old_ivars->ivar_list[ivar];
        struct objc_ivar *new_ivar = &new_ivars->ivar_list[ivar];

        new_ivar->ivar_name = old_ivar->ivar_name;
        new_ivar->ivar_type = old_ivar->ivar_type;

        new_ivar->ivar_offset = offset;
        offset += old_ivar->ivar_offset;
    }

    new_class->instance_size = offset;
    new_class->ivars = new_ivars;
}

jclass joc_ToClass(JNIEnv *env, Class value) {
    _assert(false);
    return NULL;
}

struct objc_method_list **copy_methods(struct objc_method_list_ *old_methods, size_t java_methods) {
    struct objc_method_list **method_list = apr_palloc(apr_pool_, 2 * sizeof(struct objc_method_list *));

    struct objc_method_list *new_methods = apr_palloc(apr_pool_,
        sizeof(struct objc_method_list_) +
        ((old_methods == NULL ? 0 : old_methods->method_count) + java_methods) * sizeof(struct objc_method)
    );

    method_list[0] = new_methods;
    method_list[1] = NULL;

    if (old_methods != NULL)
        memcpy(new_methods, old_methods, sizeof(struct objc_method_list_) + old_methods->method_count * sizeof(struct objc_method));
    else {
        new_methods->obsolete = NULL;
        new_methods->method_count = 0;
    }

    return method_list;
}

void (*java_instance_call(struct sig_type *type, JNIEnv *env)) {
    switch (type->primitive) {
        case typename_P: return FFI_FN((*env)->CallObjectMethod);
        case union_P: return FFI_FN((*env)->CallObjectMethod);
        case string_P: return FFI_FN((*env)->CallObjectMethod);
        case selector_P: return FFI_FN((*env)->CallObjectMethod);
        case object_P: return FFI_FN((*env)->CallObjectMethod);
        case boolean_P: return FFI_FN((*env)->CallBooleanMethod);
        case uchar_P: return FFI_FN((*env)->CallByteMethod);
        case uint_P: return FFI_FN((*env)->CallIntMethod);
        case ulong_P: return FFI_FN((*env)->CallIntMethod);
        case ulonglong_P: return FFI_FN((*env)->CallLongMethod);
        case ushort_P: return FFI_FN((*env)->CallShortMethod);
        case array_P: return FFI_FN((*env)->CallObjectMethod);
        case pointer_P: return FFI_FN((*env)->CallObjectMethod);
        case bit_P: _assert(false); break;
        case char_P: return FFI_FN((*env)->CallByteMethod);
        case double_P: return FFI_FN((*env)->CallDoubleMethod);
        case float_P: return FFI_FN((*env)->CallFloatMethod);
        case int_P: return FFI_FN((*env)->CallIntMethod);
        case long_P: return FFI_FN((*env)->CallIntMethod);
        case longlong_P: return FFI_FN((*env)->CallLongMethod);
        case short_P: return FFI_FN((*env)->CallShortMethod);
        case void_P: return FFI_FN((*env)->CallVoidMethod);
        case struct_P: return FFI_FN((*env)->CallObjectMethod);
        default: _assert(false); break;
    }
}

void (*java_static_call(struct sig_type *type, JNIEnv *env)) {
    switch (type->primitive) {
        case typename_P: return FFI_FN((*env)->CallStaticObjectMethod);
        case union_P: return FFI_FN((*env)->CallStaticObjectMethod);
        case string_P: return FFI_FN((*env)->CallStaticObjectMethod);
        case selector_P: return FFI_FN((*env)->CallStaticObjectMethod);
        case object_P: return FFI_FN((*env)->CallStaticObjectMethod);
        case boolean_P: return FFI_FN((*env)->CallStaticBooleanMethod);
        case uchar_P: return FFI_FN((*env)->CallStaticByteMethod);
        case uint_P: return FFI_FN((*env)->CallStaticIntMethod);
        case ulong_P: return FFI_FN((*env)->CallStaticIntMethod);
        case ulonglong_P: return FFI_FN((*env)->CallStaticLongMethod);
        case ushort_P: return FFI_FN((*env)->CallStaticShortMethod);
        case array_P: return FFI_FN((*env)->CallStaticObjectMethod);
        case pointer_P: return FFI_FN((*env)->CallStaticObjectMethod);
        case bit_P: _assert(false); break;
        case char_P: return FFI_FN((*env)->CallStaticByteMethod);
        case double_P: return FFI_FN((*env)->CallStaticDoubleMethod);
        case float_P: return FFI_FN((*env)->CallStaticFloatMethod);
        case int_P: return FFI_FN((*env)->CallStaticIntMethod);
        case long_P: return FFI_FN((*env)->CallStaticIntMethod);
        case longlong_P: return FFI_FN((*env)->CallStaticLongMethod);
        case short_P: return FFI_FN((*env)->CallStaticShortMethod);
        case void_P: return FFI_FN((*env)->CallStaticVoidMethod);
        case struct_P: return FFI_FN((*env)->CallStaticObjectMethod);
        default: _assert(false); break;
    }
}

Class joc_FromClass(JNIEnv *env, jclass value) {
    if (value == NULL)
        return NULL;

    jocSelect_(meta_methods[0]);
    jocSelect_(instance_methods[0]);

    bool is_objc_class;
    char *objc_class_name; {
        jstring java_class_name = (*(*env)->CallObjectMethod)(env, value, java_lang_Class_getName);
        const char *c_class_name = (*(*env)->GetStringUTFChars)(env, java_class_name, NULL);
        is_objc_class = strncmp(c_class_name, "obc.", 4) == 0;
        objc_class_name = joc_FromName(c_class_name);
        (*(*env)->ReleaseStringUTFChars)(env, java_class_name, c_class_name);
        (*(*env)->DeleteLocalRef)(env, java_class_name);
    }

    Class objc_class = objc_getClass(objc_class_name);
    if (objc_class != NULL) {
        free(objc_class_name);
        return [[objc_class retain] autorelease];
    }

    if (is_objc_class) {
        printf("objc_getClass(\"%s\")\n", objc_class_name);
        _assert(false);
    }

    struct objc_class *instance_super; {
        jclass java_super = (*(*env)->CallObjectMethod)(env, value, java_lang_Class_getSuperclass);
        instance_super = joc_FromClass(env, java_super);
        _assert(instance_super != NULL);
        (*(*env)->DeleteLocalRef)(env, java_super);
    }

    struct objc_class *meta_super = instance_super->isa;
    struct objc_class *meta_isa = meta_super->isa;

    struct objc_class *meta_class = apr_palloc(apr_pool_, meta_isa->instance_size);
    struct objc_class *instance_class = apr_palloc(apr_pool_, meta_super->instance_size + sizeof(struct meta_strap));
    struct meta_strap *meta_strap = (struct meta_strap *) ((uint8_t *) instance_class + meta_super->instance_size);

    meta_strap->joc_env = env;

    meta_strap->joc_class = (*(*env)->NewGlobalRef)(env, value);
    _assert(meta_strap->joc_class != NULL);

    meta_strap->joc_method = (*(*env)->GetMethodID)(env, value, "<init>",  "()V");
    if (meta_strap->joc_method == NULL)
        (*(*env)->ExceptionClear)(env);

    meta_class->isa = meta_isa;
    meta_class->super_class = meta_super;
    meta_class->name = objc_class_name;
    meta_class->version = 0;
    meta_class->info = CLS_META;
    meta_class->cache = NULL;
    meta_class->protocols = NULL;

    instance_class->isa = meta_class;
    instance_class->super_class = instance_super;
    instance_class->name = objc_class_name;
    instance_class->version = 0;
    instance_class->info = CLS_CLASS;
    instance_class->cache = NULL;
    instance_class->protocols = NULL;

    //printf("instance_class:%p ->info:%.8lx meta_class:%p\n", instance_class, instance_class->info, meta_class);

    jobjectArray java_methods = (*(*env)->CallObjectMethod)(env, value, java_lang_Class_getDeclaredMethods);
    jsize method_count = (*(*env)->GetArrayLength)(env, java_methods);

    if ((instance_super->info & CLS_JAVA_CLASS) == 0) {
        meta_class->methodLists = copy_methods(&meta_method, method_count);
        instance_class->methodLists = copy_methods(&instance_method, method_count);
    } else {
        meta_class->methodLists = copy_methods(NULL, method_count);
        instance_class->methodLists = copy_methods(NULL, method_count);
    }

    for (jsize method_index = 0; method_index != method_count; ++method_index) {
        jobject java_method = (*(*env)->GetObjectArrayElement)(env, java_methods, method_index);
        jobject message = (*(*env)->CallObjectMethod)(env, java_method, java_lang_reflect_Method_getAnnotation, joc_Message);

        if (message == NULL)
            goto free_method;

        jstring message_jname = (*(*env)->CallObjectMethod)(env, message, joc_Message_name);
        jstring message_jtypes = (*(*env)->CallObjectMethod)(env, message, joc_Message_types);

        bool converted;
        if ((*(*env)->GetStringLength)(env, message_jname) != 0)
            converted = false;
        else {
            converted = true;
            (*(*env)->DeleteLocalRef)(env, message_jname);
            message_jname = (*(*env)->CallObjectMethod)(env, java_method, java_lang_reflect_Method_getName);
        }

        const char *message_cname = (*(*env)->GetStringUTFChars)(env, message_jname, NULL);

        const char *name;
        char *converting;

        if (!converted) {
            converting = NULL;
            name = message_cname;
        } else {
            converting = strdup(message_cname);
            for (size_t size = strlen(converting), i = 0; i != size; ++i)
                if (converting[i] == '$')
                    converting[i] = ':';
            name = converting;
        }

        if ((*(*env)->GetStringLength)(env, message_jtypes) == 0)
            message_jtypes = (*(*env)->CallStaticObjectMethod)(env, joc_Runtime, joc_Runtime_getSignature, java_method);

        const char *message_ctypes = (*(*env)->GetStringUTFChars)(env, message_jtypes, NULL);

        jint method_modifiers = (*(*env)->CallIntMethod)(env, java_method, java_lang_reflect_Method_getModifiers);
        jboolean method_static = (*(*env)->CallStaticBooleanMethod)(env, java_lang_reflect_Modifier, java_lang_reflect_Modifier_isStatic, method_modifiers);

        struct objc_method_list *method_list = (method_static ? meta_class : instance_class)->methodLists[0];
        struct objc_method *objc_method = &method_list->method_list[method_list->method_count++];

        struct sig_signature *signature = apr_palloc(apr_pool_, sizeof(struct sig_signature));
        sig_parse(apr_pool_, signature, message_ctypes);

        ffi_type **java_types = apr_palloc(apr_pool_, (signature->count - 1 + 3) * sizeof(ffi_type));
        java_types[0] = &ffi_type_pointer;
        java_types[1] = &ffi_type_pointer;
        java_types[2] = &ffi_type_pointer;

        ffi_cif *java_cif = apr_palloc(apr_pool_, sizeof(ffi_cif));
        sig_ffi_cif(apr_pool_, &sig_java_ffi_type, signature, java_cif, java_types, 2, 3);

        ffi_type **objc_types = apr_palloc(apr_pool_, (signature->count - 1) * sizeof(ffi_type));
        ffi_cif *objc_cif = apr_palloc(apr_pool_, sizeof(ffi_cif));
        sig_ffi_cif(apr_pool_, &sig_objc_ffi_type, signature, objc_cif, objc_types, 0, 0);

        /* XXX: this memory page should be pooled with the other ones for the sake of even reasonable efficiency */
        ffi_closure *closure;
        _syscall(closure = (ffi_closure *) mmap(NULL, sizeof(ffi_closure), PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE, -1, 0));

        struct joc_userdata *userdata = apr_palloc(apr_pool_, sizeof(struct joc_userdata));
        userdata->signature = signature;
        userdata->env = env;
        userdata->method = (*(*env)->FromReflectedMethod)(env, java_method);
        userdata->cif = java_cif;
        userdata->fn = (*(method_static ? &java_static_call : &java_instance_call))(signature->elements[0].type, env);

        ffi_status status = ffi_prep_closure(closure, objc_cif, &joc_closure, userdata);
        _assert(status == FFI_OK);

        _syscall(mprotect(closure, sizeof(*closure), PROT_READ | PROT_EXEC));

        objc_method->method_name = sel_registerName(name);
        objc_method->method_types = strdup(message_ctypes);
        objc_method->method_imp = (IMP) closure;

        free(converting);

        (*(*env)->ReleaseStringUTFChars)(env, message_jtypes, message_ctypes);
        (*(*env)->DeleteLocalRef)(env, message_jtypes);

        (*(*env)->ReleaseStringUTFChars)(env, message_jname, message_cname);
        (*(*env)->DeleteLocalRef)(env, message_jname);

        (*(*env)->DeleteLocalRef)(env, message);
      free_method:
        (*(*env)->DeleteLocalRef)(env, java_method);
    }

    copy_ivars(meta_class, &meta_ivars_);

    if ((instance_super->info & CLS_JAVA_CLASS) == 0)
        copy_ivars(instance_class, &instance_ivars_);
    else {
        instance_class->instance_size = instance_super->instance_size;
        instance_class->ivars = (struct objc_ivar_list *) (void *) &null_ivars_;
    }

    (*(*env)->DeleteLocalRef)(env, java_methods);

    objc_addClass(instance_class);
    instance_class->info |= CLS_JAVA_CLASS;
    meta_class->info |= CLS_JAVA_CLASS;

    return [[instance_class retain] autorelease];
}

char *joc_ToName(const char *cname) {
    char *jname;
    if (strchr(cname, '.') == NULL)
        asprintf(&jname, "obc.%s", cname);
    else
        jname = strdup(cname);

    size_t size = strlen(jname);
    for (size_t i = 0; i != size; ++i)
        if (jname[i] == ':')
            jname[i] = '$';

    return jname;
}

char *joc_FromName(const char *jname) {
    char *cname;
    if (strncmp(jname, "obc.", 4) == 0)
        cname = strdup(jname + 4);
    else
        cname = strdup(jname);
    return cname;
}

jobject Java_joc_Runtime_toString(JNIEnv *env, jclass _this, jobject object) {
    return joc_ToString(env, joc_FromObject(env, object));
}

void Java_joc_Runtime_allocate(JNIEnv *env, jclass _this, jobject java_object) {
    jclass java_class = (*(*env)->GetObjectClass)(env, java_object);
    Class objc_class = joc_FromClass(env, java_class);
    (*(*env)->DeleteLocalRef)(env, java_class);

    //printf("objc_class:%p ->info:%.8lx\n", objc_class, objc_class->info);

    id objc_object;
    jboolean incorrect;

    if ((objc_class->info & CLS_JAVA_CLASS) != 0) {
        objc_object = [objc_class allocWithObject: java_object];
        incorrect = false;
    } else {
        objc_object = [objc_class alloc];
        incorrect = objc_object->isa != objc_class;
    }

    (*(*env)->SetLongField)(env, java_object, joc_Strap_address, (jlong) (uintptr_t) objc_object);
    (*(*env)->SetBooleanField)(env, java_object, joc_Strap_incorrect, incorrect);
}

void (*objc_fn(ffi_type *ffi_type))() {
    return ffi_type->type == FFI_TYPE_STRUCT && (
        ffi_type->size > OBJC_MAX_STRUCT_BY_VALUE ||
        struct_forward_array[ffi_type->size] != 0
    ) ? FFI_FN(&objc_msgSendSuper_stret) : FFI_FN(&objc_msgSendSuper);
}

jlong Java_joc_Runtime_dlopen(JNIEnv *env, jclass _this, jstring path) {
    const char *cpath = (*(*env)->GetStringUTFChars)(env, path, NULL);
    printf("[%s]\n", cpath);
    void *handle = dlopen(cpath, RTLD_LAZY | RTLD_GLOBAL);
    (*env)->ReleaseStringUTFChars(env, path, cpath);
    return (jlong) (uintptr_t) handle;
}

jobject Java_joc_Runtime_msgSend(JNIEnv *env, jclass _this, jobject object, jclass _class, jstring name, jobjectArray arguments) {
    id cobject = joc_FromObject(env, object);

    Class cclass = joc_FromClass(env, _class);
    if (cclass == NULL)
        cclass = [cobject class];

    SEL selector; {
        const char *cname = (*(*env)->GetStringUTFChars)(env, name, NULL);
        selector = sel_getUid(cname);
        (*env)->ReleaseStringUTFChars(env, name, cname);
    }

    Method method;
    if ((cobject->isa->info & CLS_META) == 0) {
        method = class_getInstanceMethod(cclass, selector);
    } else {
        method = class_getClassMethod(cobject, selector);
        cclass = cclass->isa;
    }

    _assert(cobject != NULL);
    struct objc_super super = {cobject, cclass};

    //printf("cobject:%p (%s) selector:%p (%s) method:%p cclass:%p cobject->isa:%p\n", cobject, cobject->isa->name, selector, sel_getName(selector), method, cclass, cobject->isa);

    apr_pool_t *pool;
    apr_pool_create(&pool, apr_pool_);

    struct sig_signature signature;
    sig_parse(pool, &signature, method->method_types);
    unsigned count = signature.count - 1;

    _assert(count >= 2);
    _assert(signature.elements[1].type->primitive == object_P);
    _assert(signature.elements[2].type->primitive == selector_P);

    ffi_cif ffi_cif;
    ffi_type *types[count];
    sig_ffi_cif(pool, &sig_objc_ffi_type, &signature, &ffi_cif, types, 0, 0);

    /* XXX: alignment */
    uint8_t data[signature.elements[0].offset];
    void *values[count];

    for (unsigned index = 0; index != count; ++index) {
        struct sig_element *element = &signature.elements[index + 1];
        values[index] = data + element->offset;

        if (index == 0)
            * (struct objc_super **) values[index] = &super;
        else if (index == 1)
            * (SEL *) values[index] = selector;
        else {
            jobject object = (*(*env)->GetObjectArrayElement)(env, arguments, index - 2);
            java_cast_from(pool, element->type, types[index], values[index], env, object);
            (*(*env)->DeleteLocalRef)(env, object);
        }
    }

    void (*fn)() = objc_fn(ffi_cif.rtype);

    /* XXX: alignment */
    uint8_t value[ffi_cif.rtype->size];

    jthrowable jerror = NULL;

    @try {
        ffi_call(&ffi_cif, fn, value, values);
    } @catch (id error) {
        jerror = joc_ToException(env, error);
    }

    //hexdump("sendMsg:value", value, sizeof(value));

    jobject result;
    if (jerror == NULL)
        result = java_cast_to(pool, signature.elements[0].type, ffi_cif.rtype, value, env);
    else {
        result = NULL;
        (*(*env)->Throw)(env, jerror);
    }

    apr_pool_destroy(pool);
    return result;
}
